from bip_utils.bip.bip32.base.bip32_base import Bip32Base
from bip_utils.bip.bip32.base.ibip32_key_derivator import IBip32KeyDerivator
from bip_utils.bip.bip32.base.ibip32_mst_key_generator import IBip32MstKeyGenerator



from bip_utils.substrate.substrate import Substrate, SubstrateCoins
from bip_utils.substrate.substrate_ex import SubstrateKeyError, SubstratePathError
from bip_utils.substrate.substrate_keys import SubstratePrivateKey, SubstratePublicKey
from bip_utils.substrate.substrate_path import SubstratePath, SubstratePathElem, SubstratePathParser



from bip_utils.monero.mnemonic.monero_entropy_generator import MoneroEntropyBitLen, MoneroEntropyGenerator
from bip_utils.monero.mnemonic.monero_mnemonic import MoneroLanguages, MoneroMnemonic, MoneroWordsNum
from bip_utils.monero.mnemonic.monero_mnemonic_decoder import MoneroMnemonicDecoder
from bip_utils.monero.mnemonic.monero_mnemonic_encoder import (
    MoneroMnemonicEncoder, MoneroMnemonicNoChecksumEncoder, MoneroMnemonicWithChecksumEncoder
)
from bip_utils.monero.mnemonic.monero_mnemonic_generator import MoneroMnemonicGenerator
from bip_utils.monero.mnemonic.monero_mnemonic_validator import MoneroMnemonicValidator
from bip_utils.monero.mnemonic.monero_seed_generator import MoneroSeedGenerator



from bip_utils.utils.misc.algo import AlgoUtils
from bip_utils.utils.misc.base32 import Base32Decoder, Base32Encoder
from bip_utils.utils.misc.bit import BitUtils
from bip_utils.utils.misc.bytes import BytesUtils
from bip_utils.utils.misc.cbor_indefinite_len_array import CborIndefiniteLenArrayDecoder, CborIndefiniteLenArrayEncoder
from bip_utils.utils.misc.data_bytes import DataBytes
from bip_utils.utils.misc.integer import IntegerUtils
from bip_utils.utils.misc.string import StringUtils



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with BIP38 utility functions."""

# Imports
from typing import Union

from bip_utils.addr import P2PKHAddr
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import DoubleSha256
from bip_utils.wif import WifPubKeyModes


# Alias for WifPubKeyModes
Bip38PubKeyModes = WifPubKeyModes


class Bip38AddrConst:
    """Class container for BIP38 address constants."""

    # Address hash length
    ADDR_HASH_LEN: int = 4


class Bip38Addr:
    """Class for BIP38 address computation."""

    @staticmethod
    def AddressHash(pub_key: Union[bytes, IPublicKey],
                    pub_key_mode: Bip38PubKeyModes) -> bytes:
        """
        Compute the address hash as specified in BIP38.

        Args:
            pub_key (bytes or IPublicKey)  : Public key bytes or object
            pub_key_mode (Bip38PubKeyModes): Public key mode

        Returns:
            bytes: Address hash

        Raises:
            TypeError: If the public key is not a Secp256k1PublicKey
            ValueError: If the public key bytes are not valid
        """

        # Compute the Bitcoin address
        address = P2PKHAddr.EncodeKey(pub_key,
                                      net_ver=CoinsConf.BitcoinMainNet.ParamByKey("p2pkh_net_ver"),
                                      pub_key_mode=pub_key_mode)
        # Take the first four bytes of SHA256(SHA256())
        return DoubleSha256.QuickDigest(address)[:Bip38AddrConst.ADDR_HASH_LEN]



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Cardano Shelley keys handling."""

# Imports
from functools import lru_cache

from bip_utils.addr import AdaShelleyAddrEncoder, AdaShelleyStakingAddrEncoder
from bip_utils.bip.bip32 import Bip32PrivateKey, Bip32PublicKey
from bip_utils.bip.conf.common import BipCoinConf


class CardanoShelleyPublicKeys:
    """
    Cardano Shelley public key class.
    It contains 2 CIP-1852 public keys (address + staking) and allows to get the Cardano Shelley address from them.
    """

    m_pub_addr_key: Bip32PublicKey
    m_pub_sk_key: Bip32PublicKey
    m_coin_conf: BipCoinConf

    def __init__(self,
                 pub_addr_key: Bip32PublicKey,
                 pub_sk_key: Bip32PublicKey,
                 coin_conf: BipCoinConf) -> None:
        """
        Construct class.

        Args:
            pub_addr_key (Bip32PublicKey object): Bip32PublicKey object (address)
            pub_sk_key (Bip32PublicKey object)  : Bip32PublicKey object (staking)
            coin_conf (BipCoinConf object)      : BipCoinConf object
        """
        self.m_pub_addr_key = pub_addr_key
        self.m_pub_sk_key = pub_sk_key
        self.m_coin_conf = coin_conf

    def AddressKey(self) -> Bip32PublicKey:
        """
        Get the address public key.

        Returns:
            Bip32PublicKey object: Bip32PublicKey object
        """
        return self.m_pub_addr_key

    def RewardKey(self) -> Bip32PublicKey:
        """
        Alias for StakingKey.

        Returns:
            Bip32PublicKey object: Bip32PublicKey object
        """
        return self.StakingKey()

    def StakingKey(self) -> Bip32PublicKey:
        """
        Get the staking address public key.

        Returns:
            Bip32PublicKey object: Bip32PublicKey object
        """
        return self.m_pub_sk_key

    def ToRewardAddress(self) -> str:
        """
        Alias for ToStakingAddress.

        Returns:
            str: Reward address string

        Raises:
            ValueError: If the public key is not correspondent to an address index level
        """
        return self.ToStakingAddress()

    @lru_cache()
    def ToStakingAddress(self) -> str:
        """
        Return the staking address correspondent to the public key.

        Returns:
            str: Staking address string

        Raises:
            ValueError: If the public key is not correspondent to an address index level
        """
        return AdaShelleyStakingAddrEncoder.EncodeKey(self.m_pub_sk_key.KeyObject(),
                                                      **self.m_coin_conf.AddrParams())

    @lru_cache()
    def ToAddress(self) -> str:
        """
        Return the address correspondent to the public key.

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not correspondent to an address index level
        """
        return AdaShelleyAddrEncoder.EncodeKey(self.m_pub_addr_key.KeyObject(),
                                               pub_skey=self.m_pub_sk_key.KeyObject(),
                                               **self.m_coin_conf.AddrParams())


class CardanoShelleyPrivateKeys:
    """
    Cardano Shelley private key class.
    It contains 2 BIP32 private keys (address + staking).
    """

    m_priv_addr_key: Bip32PrivateKey
    m_priv_sk_key: Bip32PrivateKey
    m_coin_conf: BipCoinConf

    def __init__(self,
                 priv_addr_key: Bip32PrivateKey,
                 priv_sk_key: Bip32PrivateKey,
                 coin_conf: BipCoinConf) -> None:
        """
        Construct class.

        Args:
            priv_addr_key (Bip32PrivateKey object): Bip32PrivateKey object (address)
            priv_sk_key (Bip32PrivateKey object)  : Bip32PrivateKey object (staking)
            coin_conf (BipCoinConf object)        : BipCoinConf object
        """
        self.m_priv_addr_key = priv_addr_key
        self.m_priv_sk_key = priv_sk_key
        self.m_coin_conf = coin_conf

    def AddressKey(self) -> Bip32PrivateKey:
        """
        Get the address private key.

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object
        """
        return self.m_priv_addr_key

    def RewardKey(self) -> Bip32PrivateKey:
        """
        Alias for StakingKey.

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object
        """
        return self.StakingKey()

    def StakingKey(self) -> Bip32PrivateKey:
        """
        Get the staking address private key.

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object
        """
        return self.m_priv_sk_key

    @lru_cache()
    def PublicKeys(self) -> CardanoShelleyPublicKeys:
        """
        Get the public keys correspondent to the private ones.

        Returns:
            CardanoShelleyPublicKeys object: CardanoShelleyPublicKeys object
        """
        return CardanoShelleyPublicKeys(self.m_priv_addr_key.PublicKey(),
                                        self.m_priv_sk_key.PublicKey(),
                                        self.m_coin_conf)



from bip_utils.bip.conf.common.bip_bitcoin_cash_conf import BipBitcoinCashConf
from bip_utils.bip.conf.common.bip_coin_conf import BipCoinConf, BipCoinFctCallsConf
from bip_utils.bip.conf.common.bip_coins import BipCoins
from bip_utils.bip.conf.common.bip_conf_const import (
    DER_PATH_HARDENED_FULL, DER_PATH_HARDENED_SHORT, DER_PATH_NON_HARDENED_FULL
)
from bip_utils.bip.conf.common.bip_litecoin_conf import BipLitecoinConf



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Substrate paths parsing and handling."""

# Imports
from __future__ import annotations

import re
from functools import lru_cache
from typing import Dict, Iterator, List, Optional, Sequence, Type, Union

from bip_utils.substrate.scale import (
    SubstrateScaleBytesEncoder, SubstrateScaleEncoderBase, SubstrateScaleU8Encoder, SubstrateScaleU16Encoder,
    SubstrateScaleU32Encoder, SubstrateScaleU64Encoder, SubstrateScaleU128Encoder, SubstrateScaleU256Encoder
)
from bip_utils.substrate.substrate_ex import SubstratePathError
from bip_utils.utils.crypto import Blake2b256


class SubstratePathConst:
    """Container for Substrate path constants."""

    # Encoded element maximum length in bytes
    ENCODED_ELEM_MAX_BYTE_LEN: int = 32
    # Regex for path
    RE_PATH: str = r"\/+[^/]+"

    # Soft path prefix
    SOFT_PATH_PREFIX: str = "/"
    # hard path prefix
    HARD_PATH_PREFIX: str = "//"

    # SCALE encoders for integers
    SCALE_INT_ENCODERS: Dict[int, Type[SubstrateScaleEncoderBase]] = {
        8: SubstrateScaleU8Encoder,
        16: SubstrateScaleU16Encoder,
        32: SubstrateScaleU32Encoder,
        64: SubstrateScaleU64Encoder,
        128: SubstrateScaleU128Encoder,
        256: SubstrateScaleU256Encoder,
    }


class SubstratePathElem:
    """
    Substrate path element.
    It represents a Substrate path element.
    """

    m_elem: str
    m_is_hard: bool

    def __init__(self,
                 elem: str) -> None:
        """
        Construct class.

        Args:
            elem (str): Path element

        Raises:
            SubstratePathError: If the path element is not valid
        """
        if not self.__IsElemValid(elem):
            raise SubstratePathError(f"Invalid path element ({elem})")

        self.m_elem = elem.replace("/", "")
        self.m_is_hard = elem.startswith(SubstratePathConst.HARD_PATH_PREFIX)

    def IsHard(self) -> bool:
        """
        Get if the element is hard.

        Returns:
            bool: True if hard, false otherwise
        """
        return self.m_is_hard

    def IsSoft(self) -> bool:
        """
        Get if the element is soft.

        Returns:
            bool: True if soft, false otherwise
        """
        return not self.IsHard()

    @lru_cache()
    def ChainCode(self) -> bytes:
        """
        Return the chain code.

        Returns:
            bytes: Chain code
        """
        return self.__ComputeChainCode()

    def ToStr(self) -> str:
        """
        Get the path element as a string.

        Returns:
            str: Path element as a string
        """
        prefix = SubstratePathConst.HARD_PATH_PREFIX if self.m_is_hard else SubstratePathConst.SOFT_PATH_PREFIX
        return prefix + self.m_elem

    def __str__(self) -> str:
        """
        Get the path element as a string.

        Returns:
            str: Path element as a string
        """
        return self.ToStr()

    def __ComputeChainCode(self) -> bytes:
        """
        Compute chain code.

        Returns:
            bytes: Chain code

        Raises:
            SubstratePathError: If path is a number bigger than 256-bit
        """

        # Integer
        if self.m_elem.isnumeric():
            bit_len = int(self.m_elem).bit_length()

            # Find the correct scale encoder
            scale_enc = None
            for min_bit_len, int_scale_enc in SubstratePathConst.SCALE_INT_ENCODERS.items():
                if bit_len <= min_bit_len:
                    scale_enc = int_scale_enc
                    break

            if scale_enc is None:
                raise SubstratePathError(f"Invalid integer bit length ({bit_len})")
        # String
        else:
            scale_enc = SubstrateScaleBytesEncoder

        # Encode element
        enc_data = scale_enc.Encode(self.m_elem)

        # Compute chain code
        max_len = SubstratePathConst.ENCODED_ELEM_MAX_BYTE_LEN
        if len(enc_data) > max_len:
            chain_code = Blake2b256.QuickDigest(enc_data)
        else:
            chain_code = enc_data.ljust(max_len, b"\x00")

        return chain_code

    @staticmethod
    def __IsElemValid(elem: str) -> bool:
        """
        Get a path element is valid.

        Args:
            elem (str): Path element

        Returns:
            bool: True if valid, false otherwise
        """
        return ((elem.startswith(SubstratePathConst.SOFT_PATH_PREFIX)
                 or elem.startswith(SubstratePathConst.HARD_PATH_PREFIX))
                and elem.rfind("/") < 2
                and len(elem.replace("/", "")) > 0)


class SubstratePath:
    """
    Substrate path.
    It represents a Substrate path.
    """

    m_elems: List[SubstratePathElem]

    def __init__(self,
                 elems: Optional[Sequence[Union[str, SubstratePathElem]]] = None) -> None:
        """
        Construct class.

        Args:
            elems (list, optional): Path elements (default: empty)

        Raises:
            SubstratePathError: If the path element is not valid
        """
        self.m_elems = ([]
                        if elems is None
                        else [SubstratePathElem(elem) if isinstance(elem, str) else elem for elem in elems])

    def AddElem(self,
                elem: Union[str, SubstratePathElem]) -> SubstratePath:
        """
        Return a new path object with the specified element added.

        Args:
            elem (str or SubstratePathElem): Path element

        Returns:
            SubstratePath object: SubstratePath object

        Raises:
            SubstratePathError: If the path element is not valid
        """
        if isinstance(elem, str):
            elem = SubstratePathElem(elem)
        return SubstratePath(self.m_elems + [elem])

    def Length(self) -> int:
        """
        Get the number of elements of the path.

        Returns:
            int: Number of elements
        """
        return len(self.m_elems)

    def ToList(self) -> List[str]:
        """
        Get the path as a list of strings.

        Returns:
            list[str]: Path as a list of strings
        """
        return [str(elem) for elem in self.m_elems]

    def ToStr(self) -> str:
        """
        Get the path as a string.

        Returns:
            str: Path as a string
        """
        return "".join(self.ToList())

    def __str__(self) -> str:
        """
        Get the path as a string.

        Returns:
            str: Path as a list of integers
        """
        return self.ToStr()

    def __getitem__(self,
                    idx: int) -> SubstratePathElem:
        """
        Get the specified element index.

        Args:
            idx (int): Element index

        Returns:
            SubstratePathElem object: SubstratePathElem object
        """
        return self.m_elems[idx]

    def __iter__(self) -> Iterator[SubstratePathElem]:
        """
        Get the iterator to the current element.

        Returns:
            Iterator object: Iterator to the current element
        """
        yield from self.m_elems


class SubstratePathParser:
    """
    Substrate path parser.
    It parses a Substrate path and returns a SubstratePath object.
    """

    @staticmethod
    def Parse(path: str) -> SubstratePath:
        """
        Parse a path and return a SubstratePath object.

        Args:
            path (str): Path

        Returns:
            SubstratePath object: SubstratePath object

        Raises:
            SubstratePathError: If the path element is not valid
        """
        if len(path) > 0 and not path.startswith("/"):
            raise SubstratePathError(f"Invalid path ({path})")

        paths = re.findall(SubstratePathConst.RE_PATH, path)
        return SubstratePath(list(paths))



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v1 mnemonic seed generation."""

# Imports
from typing import Optional, Union

from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic import ElectrumV1Languages
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_decoder import ElectrumV1MnemonicDecoder
from bip_utils.utils.crypto import Sha256
from bip_utils.utils.misc import AlgoUtils, BytesUtils
from bip_utils.utils.mnemonic import Mnemonic


class ElectrumV1SeedGeneratorConst:
    """Class container for Electrum v1 seed generator constants."""

    # Number of hash iteration
    HASH_ITR_NUM: int = 10**5


class ElectrumV1SeedGenerator:
    """
    Electrum seed generator class (v1).
    It generates the seed from a mnemonic.
    """

    m_seed: bytes

    def __init__(self,
                 mnemonic: Union[str, Mnemonic],
                 lang: Optional[ElectrumV1Languages] = ElectrumV1Languages.ENGLISH) -> None:
        """
        Construct class.
        Language is set to English by default because Electrum v1 mnemonic only support one language,
        so it's useless (and slower) to automatically detect the language.

        Args:
            mnemonic (str or Mnemonic object)   : Mnemonic
            lang (ElectrumV1Languages, optional): Language, None for automatic detection

        Raises:
            ValueError: If the mnemonic is not valid
        """
        entropy_bytes = ElectrumV1MnemonicDecoder(lang).Decode(mnemonic)
        # Compute the seed only once
        self.m_seed = self.__GenerateSeed(entropy_bytes)

    def Generate(self) -> bytes:
        """
        Generate seed.
        There is no really need of this method, since the seed is always the same, but it's
        kept in this way to have the same usage of Bip39/Substrate seed generator
        (i.e. ElectrumV1SeedGenerator(mnemonic).Generate() ).

        Returns:
            bytes: Generated seed
        """
        return self.m_seed

    @staticmethod
    def __GenerateSeed(entropy_bytes: bytes) -> bytes:
        """
        Generate seed from entropy bytes.

        Args:
            entropy_bytes (bytes): Entropy bytes

        Returns:
            bytes: Generated seed
        """
        entropy_hex = AlgoUtils.Encode(BytesUtils.ToHexString(entropy_bytes))
        h = entropy_hex
        for _ in range(ElectrumV1SeedGeneratorConst.HASH_ITR_NUM):
            h = Sha256.QuickDigest(h + entropy_hex)
        return h



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for ed25519-blake2b keys."""

# Imports
from typing import Any

import ed25519_blake2b

from bip_utils.ecc.common.ikeys import IPrivateKey, IPublicKey
from bip_utils.ecc.common.ipoint import IPoint
from bip_utils.ecc.curve.elliptic_curve_types import EllipticCurveTypes
from bip_utils.ecc.ed25519.ed25519_keys import Ed25519KeysConst, Ed25519PublicKey
from bip_utils.ecc.ed25519.lib import ed25519_lib
from bip_utils.ecc.ed25519_blake2b.ed25519_blake2b_point import Ed25519Blake2bPoint
from bip_utils.utils.misc import BytesUtils, DataBytes


class Ed25519Blake2bPublicKey(IPublicKey):
    """Ed25519-Blake2b public key class."""

    m_ver_key: ed25519_blake2b.VerifyingKey

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes) -> IPublicKey:
        """
        Construct class from key bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPublicKey: IPublicKey object

        Raises:
            ValueError: If key bytes are not valid
        """

        # Remove the 0x00 prefix if present
        if (len(key_bytes) == Ed25519KeysConst.PUB_KEY_BYTE_LEN + len(Ed25519KeysConst.PUB_KEY_PREFIX)
                and key_bytes[0] == BytesUtils.ToInteger(Ed25519KeysConst.PUB_KEY_PREFIX)):
            key_bytes = key_bytes[1:]
        # The library does not raise any exception in case of length error
        elif len(key_bytes) != Ed25519KeysConst.PUB_KEY_BYTE_LEN:
            raise ValueError("Invalid public key bytes")

        # The library doesn't check if the point lies on curve
        if not ed25519_lib.point_is_on_curve(key_bytes):
            raise ValueError("Invalid public key bytes")

        return cls(ed25519_blake2b.VerifyingKey(key_bytes))

    @classmethod
    def FromPoint(cls,
                  key_point: IPoint) -> IPublicKey:
        """
        Construct class from key point.

        Args:
            key_point (IPoint object): Key point

        Returns:
            IPublicKey: IPublicKey object

        Raises:
            ValueError: If key point is not valid
        """
        return cls.FromBytes(key_point.RawEncoded().ToBytes())

    def __init__(self,
                 key_obj: ed25519_blake2b.VerifyingKey) -> None:
        """
        Construct class from key object.

        Args:
            key_obj (ed25519_blake2b.VerifyingKey): Key object
        """
        self.m_ver_key = key_obj

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_BLAKE2B

    @staticmethod
    def CompressedLength() -> int:
        """
        Get the compressed key length.

        Returns:
           int: Compressed key length
        """
        return Ed25519PublicKey.CompressedLength()

    @staticmethod
    def UncompressedLength() -> int:
        """
        Get the uncompressed key length.

        Returns:
           int: Uncompressed key length
        """
        return Ed25519PublicKey.UncompressedLength()

    def UnderlyingObject(self) -> Any:
        """
        Get the underlying object.

        Returns:
           Any: Underlying object
        """
        return self.m_ver_key

    def RawCompressed(self) -> DataBytes:
        """
        Return raw compressed public key.

        Returns:
            DataBytes object: DataBytes object
        """
        return DataBytes(Ed25519KeysConst.PUB_KEY_PREFIX + self.m_ver_key.to_bytes())

    def RawUncompressed(self) -> DataBytes:
        """
        Return raw uncompressed public key.

        Returns:
            DataBytes object: DataBytes object
        """

        # Same as compressed
        return self.RawCompressed()

    def Point(self) -> IPoint:
        """
        Get public key point.

        Returns:
            IPoint object: IPoint object
        """
        return Ed25519Blake2bPoint(self.m_ver_key.to_bytes())


class Ed25519Blake2bPrivateKey(IPrivateKey):
    """Ed25519-Blake2b private key class."""

    m_sign_key: ed25519_blake2b.SigningKey

    @classmethod
    def FromBytes(cls,
                  key_bytes: bytes) -> IPrivateKey:
        """
        Construct class from key bytes.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            IPrivateKey: IPrivateKey object

        Raises:
            ValueError: If key bytes are not valid
        """
        try:
            return cls(ed25519_blake2b.SigningKey(key_bytes))
        except ValueError as ex:
            raise ValueError("Invalid private key bytes") from ex

    def __init__(self,
                 key_obj: ed25519_blake2b.SigningKey) -> None:
        """
        Construct class from key object.

        Args:
            key_obj (ed25519_blake2b.SigningKey): Key object
        """
        self.m_sign_key = key_obj

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.ED25519_BLAKE2B

    @staticmethod
    def Length() -> int:
        """
        Get the key length.

        Returns:
           int: Key length
        """
        return Ed25519KeysConst.PRIV_KEY_BYTE_LEN

    def UnderlyingObject(self) -> Any:
        """
        Get the underlying object.

        Returns:
           Any: Underlying object
        """
        return self.m_sign_key

    def Raw(self) -> DataBytes:
        """
        Return raw private key.

        Returns:
            DataBytes object: DataBytes object
        """
        return DataBytes(self.m_sign_key.to_bytes())

    def PublicKey(self) -> IPublicKey:
        """
        Get the public key correspondent to the private one.

        Returns:
            IPublicKey object: IPublicKey object
        """
        return Ed25519Blake2bPublicKey(self.m_sign_key.get_verifying_key())



import os
import re

import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install


# Load long description
def load_long_description(desc_file):
    return open(desc_file, encoding='utf-8').read()


# Load keywords
def load_keywords(keywords_file):
    with open(keywords_file, "r") as fin:
        return ", ".join([line for line in map(str.strip, fin.read().splitlines())
                          if len(line) > 0 and not line.startswith("#")])


# Load version
def load_version(*path_parts):
    version_file = os.path.join(*path_parts)
    version_line = open(os.path.join(*path_parts)).read().rstrip()
    vre = re.compile(r'__version__: str = "([^"]+)"')
    matches = vre.findall(version_line)

    if matches and len(matches) > 0:
        return matches[0]

    raise RuntimeError(f"Cannot find version string in {version_file}")


# Load requirements
def load_requirements(req_file):
    with open(req_file, "r") as fin:
        return [line for line in map(str.strip, fin.read().splitlines())
                if len(line) > 0 and not line.startswith("#")]


# Command base class
class CommandBase(object):
    conf_file = "bip_utils/ecc/conf.py"
    conf_str = "USE_COINCURVE: bool = "

    user_options = [
        ("coincurve=", None, "1 to use coincurve library for secp256k1, 0 for using ecdsa library for secp256k1"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.coincurve = 1

    def finalize_options(self):
        self.coincurve = self.__validate_and_get_coincurve_param()
        super().finalize_options()

    def run(self):
        print(f"Coincurve: {'enabled' if self.coincurve else 'disabled'}")
        self.__write_conf_file()
        super().run()

    def __validate_and_get_coincurve_param(self):
        try:
            param = int(self.coincurve)
        except ValueError:
            raise ValueError(f"Invalid coincurve option value {self.coincurve}")

        if param not in (0, 1):
            raise ValueError(f"Invalid coincurve option value {self.coincurve}")

        return param

    def __write_conf_file(self):
        with open(self.conf_file, "r") as fin:
            file_content = fin.read()

        coincurve_str = "True" if self.coincurve else "False"

        conf_idx = file_content.index(self.conf_str)
        nl_idx = file_content.index("\n", conf_idx)

        # Re-write the py file with the updated string
        # Probably not the best solution, but I'm not very familiar with install options
        file_content = file_content.replace(file_content[conf_idx:nl_idx], self.conf_str + coincurve_str, 1)

        with open(self.conf_file, "w") as fout:
            fout.write(file_content)


# Install command class
class InstallCommand(CommandBase, install):
    user_options = getattr(install, 'user_options', []) + CommandBase.user_options


# Develop command class
class DevelopCommand(CommandBase, develop):
    user_options = getattr(develop, 'user_options', []) + CommandBase.user_options


# Load version
version = load_version("bip_utils", "_version.py")

# Setup configuration
setuptools.setup(
    name="open-aea-bip-utils",
    version=version,
    author="Emanuele Bellocchia",
    author_email="ebellocchia@gmail.com",
    maintainer="Emanuele Bellocchia",
    maintainer_email="ebellocchia@gmail.com",
    description="Generation of mnemonics, seeds, private/public keys and addresses for different types of cryptocurrencies",
    long_description=load_long_description("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/ebellocchia/bip_utils",
    download_url="https://github.com/ebellocchia/bip_utils/archive/v%s.tar.gz" % version,
    license="MIT",
    test_suite="tests",
    cmdclass={
        "install": InstallCommand,
        "develop": DevelopCommand,
    },
    install_requires=load_requirements("requirements.txt"),
    extras_require={
        ":python_version == '3.7'": [
            "typing_extensions",
        ],
        "develop": load_requirements("requirements-dev.txt"),
    },
    packages=setuptools.find_packages(include=["bip_utils*"],exclude=["*tests*"]),
    package_data={
        "bip_utils": [
            # BIP39
            "bip/bip39/wordlist/english.txt",
            "bip/bip39/wordlist/italian.txt",
            "bip/bip39/wordlist/french.txt",
            "bip/bip39/wordlist/spanish.txt",
            "bip/bip39/wordlist/portuguese.txt",
            "bip/bip39/wordlist/czech.txt",
            "bip/bip39/wordlist/chinese_simplified.txt",
            "bip/bip39/wordlist/chinese_traditional.txt",
            "bip/bip39/wordlist/korean.txt",
            # Electrum
            "electrum/mnemonic_v1/wordlist/english.txt",
            # Monero
            "monero/mnemonic/wordlist/chinese_simplified.txt",
            "monero/mnemonic/wordlist/dutch.txt",
            "monero/mnemonic/wordlist/english.txt",
            "monero/mnemonic/wordlist/french.txt",
            "monero/mnemonic/wordlist/german.txt",
            "monero/mnemonic/wordlist/italian.txt",
            "monero/mnemonic/wordlist/japanese.txt",
            "monero/mnemonic/wordlist/portuguese.txt",
            "monero/mnemonic/wordlist/russian.txt",
            "monero/mnemonic/wordlist/spanish.txt",
        ]
    },
    keywords=load_keywords("keywords.txt"),
    platforms=["any"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.7",
)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Electrum v1 mnemonic utility classes."""

# Imports
import os
from typing import Tuple

from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic import ElectrumV1Languages, ElectrumV1MnemonicConst
from bip_utils.utils.mnemonic import (
    Mnemonic, MnemonicLanguages, MnemonicWordsList, MnemonicWordsListFinderBase, MnemonicWordsListGetterBase
)


class ElectrumV1WordsListGetter(MnemonicWordsListGetterBase):
    """
    Electrum words list getter class (v1).
    It allows to get words list by language so that they are loaded from file only once per language.
    """

    def GetByLanguage(self,
                      lang: MnemonicLanguages) -> MnemonicWordsList:
        """
        Get words list by language.
        Words list of a specific language are loaded from file only the first time they are requested.

        Args:
            lang (MnemonicLanguages): Language

        Returns:
            MnemonicWordsList object: MnemonicWordsList object

        Raises:
            TypeError: If the language is not a Bip39Languages enum
            ValueError: If loaded words list is not valid
        """
        if not isinstance(lang, ElectrumV1Languages):
            raise TypeError("Language is not an enumerative of Bip39Languages")

        return self._LoadWordsList(lang,
                                   self.__GetLanguageFile(lang),
                                   ElectrumV1MnemonicConst.WORDS_LIST_NUM)

    @staticmethod
    def __GetLanguageFile(lang: MnemonicLanguages) -> str:
        """
        Get the specified language file name.

        Args:
            lang (Bip39Languages): Language

        Returns:
            str: Language file name
        """
        return os.path.join(os.path.dirname(__file__),
                            ElectrumV1MnemonicConst.LANGUAGE_FILES[lang])


class ElectrumV1WordsListFinder(MnemonicWordsListFinderBase):
    """
    Electrum words list finder class (v1).
    It automatically finds the correct words list from a mnemonic.
    """

    @classmethod
    def FindLanguage(cls,
                     mnemonic: Mnemonic) -> Tuple[MnemonicWordsList, MnemonicLanguages]:
        """
        Automatically find the language of the specified mnemonic and get the correct MnemonicWordsList class for it.

        Args:
            mnemonic (Mnemonic object): Mnemonic object

        Returns:
           tuple[MnemonicWordsList, MnemonicLanguages]: MnemonicWordsList object (index 0), mnemonic language (index 1)

        Raises:
            ValueError: If the mnemonic language cannot be found
        """
        return cls._FindLanguageGeneric(mnemonic, ElectrumV1Languages, ElectrumV1WordsListGetter)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Substrate SCALE encoding for compact unsigned integers."""

# Imports
from typing import Any

from bip_utils.substrate.scale.substrate_scale_enc_base import SubstrateScaleEncoderBase
from bip_utils.utils.misc import IntegerUtils


class SubstrateScaleCUintEncoderConst:
    """Class container for Substrate SCALE encoding for compact unsigned integers constants."""

    # Maximum values
    SINGLE_BYTE_MODE_MAX_VAL: int = 2**6 - 1
    TWO_BYTE_MODE_MAX_VAL: int = 2**14 - 1
    FOUR_BYTE_MODE_MAX_VAL: int = 2**30 - 1
    BIG_INTEGER_MODE_MAX_VAL: int = 2**536 - 1


class SubstrateScaleCUintEncoder(SubstrateScaleEncoderBase):
    """Substrate SCALE encoding for compact unsigned integers."""

    @classmethod
    def Encode(cls,
               value: Any) -> bytes:
        """
        Encode the specified value to bytes.

        Args:
            value (any): Value to be encoded

        Returns:
            bytes: Encoded value
        """

        # Single-byte mode
        if value <= SubstrateScaleCUintEncoderConst.SINGLE_BYTE_MODE_MAX_VAL:
            return IntegerUtils.ToBytes(value << 2, 1, endianness="little")
        # Two-byte mode
        if value <= SubstrateScaleCUintEncoderConst.TWO_BYTE_MODE_MAX_VAL:
            return IntegerUtils.ToBytes((value << 2) | 0b01, 2, endianness="little")
        # Four-byte mode
        if value <= SubstrateScaleCUintEncoderConst.FOUR_BYTE_MODE_MAX_VAL:
            return IntegerUtils.ToBytes((value << 2) | 0b10, 4, endianness="little")
        # Big-integer mode
        if value <= SubstrateScaleCUintEncoderConst.BIG_INTEGER_MODE_MAX_VAL:
            value_bytes = IntegerUtils.ToBytes(value, endianness="little")
            len_bytes = IntegerUtils.ToBytes((len(value_bytes) - 4 << 2) | 0b11, 1, endianness="little")
            return len_bytes + value_bytes

        raise ValueError(f"Out of range integer value ({value})")



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Zilliqa address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.bech32 import Bech32ChecksumError, Bech32Decoder, Bech32Encoder
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import Sha256


class ZilAddrConst:
    """Class container for Zilliqa address constants."""

    # SHA-256 length in bytes
    SHA256_BYTE_LEN: int = 20


class ZilAddrDecoder(IAddrDecoder):
    """
    Zilliqa address decoder class.
    It allows the Zilliqa address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Zilliqa address to bytes.

        Args:
            addr (str): Address string
            **kwargs  : Not used

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """
        try:
            addr_dec_bytes = Bech32Decoder.Decode(CoinsConf.Zilliqa.ParamByKey("addr_hrp"),
                                                  addr)
        except Bech32ChecksumError as ex:
            raise ValueError("Invalid bech32 checksum") from ex

        AddrDecUtils.ValidateLength(addr_dec_bytes, ZilAddrConst.SHA256_BYTE_LEN)
        return addr_dec_bytes


class ZilAddrEncoder(IAddrEncoder):
    """
    Zilliqa address encoder class.
    It allows the Zilliqa address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Zilliqa address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object
            **kwargs                     : Not used

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1
        """
        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)

        key_hash = Sha256.QuickDigest(pub_key_obj.RawCompressed().ToBytes())
        return Bech32Encoder.Encode(CoinsConf.Zilliqa.ParamByKey("addr_hrp"),
                                    key_hash[-ZilAddrConst.SHA256_BYTE_LEN:])


# Deprecated: only for compatibility, Encoder class shall be used instead
ZilAddr = ZilAddrEncoder



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Monero subaddress computation."""

# Imports
from typing import Optional, Tuple

from bip_utils.addr import XmrAddrEncoder
from bip_utils.ecc import Ed25519Monero, Ed25519Utils
from bip_utils.monero.monero_keys import MoneroPrivateKey, MoneroPublicKey
from bip_utils.utils.crypto import Kekkak256
from bip_utils.utils.misc import IntegerUtils


class MoneroSubaddressConst:
    """Class container for Monero subaddress constants."""

    # Subaddress prefix
    SUBADDR_PREFIX: bytes = b"SubAddr\x00"
    # Subaddress maximum index
    SUBADDR_MAX_IDX: int = 2**32 - 1
    # Subaddress index length in byte
    SUBADDR_IDX_BYTE_LEN: int = 4


class MoneroSubaddress:
    """Monero subaddress class. It allows to compute Monero subaddresses."""

    m_priv_vkey: MoneroPrivateKey
    m_pub_skey: MoneroPublicKey
    m_pub_vkey: MoneroPublicKey

    def __init__(self,
                 priv_vkey: MoneroPrivateKey,
                 pub_skey: MoneroPublicKey,
                 pub_vkey: Optional[MoneroPublicKey] = None) -> None:
        """
        Construct class.

        Args:
            priv_vkey (MoneroPrivateKey)        : Private view key
            pub_skey (MoneroPublicKey)          : Public spend key
            pub_vkey (MoneroPublicKey, optional): Public view key (if None, it'll be computed from the private one)
        """
        self.m_priv_vkey = priv_vkey
        self.m_pub_skey = pub_skey
        self.m_pub_vkey = pub_vkey if pub_vkey is not None else priv_vkey.PublicKey()

    def ComputeKeys(self,
                    minor_idx: int,
                    major_idx: int) -> Tuple[MoneroPublicKey, MoneroPublicKey]:
        """
        Compute the public keys of the specified subaddress.

        Args:
            minor_idx (int): Minor index (i.e. subaddress index)
            major_idx (int): Major index (i.e. account index)

        Returns:
            tuple[MoneroPublicKey, MoneroPublicKey]: Computed public spend key (index 0) and public view key (index 1)

        Raises:
            ValueError: If one of the indexes is not valid
        """
        if minor_idx < 0 or minor_idx > MoneroSubaddressConst.SUBADDR_MAX_IDX:
            raise ValueError(f"Invalid minor index ({minor_idx})")
        if major_idx < 0 or major_idx > MoneroSubaddressConst.SUBADDR_MAX_IDX:
            raise ValueError(f"Invalid major index ({major_idx})")

        # Subaddress 0,0 is the primary address
        if minor_idx == 0 and major_idx == 0:
            return self.m_pub_skey, self.m_pub_vkey

        # Convert indexes to bytes
        major_idx_bytes = IntegerUtils.ToBytes(major_idx,
                                               bytes_num=MoneroSubaddressConst.SUBADDR_IDX_BYTE_LEN,
                                               endianness="little")
        minor_idx_bytes = IntegerUtils.ToBytes(minor_idx,
                                               bytes_num=MoneroSubaddressConst.SUBADDR_IDX_BYTE_LEN,
                                               endianness="little")

        # m = Kekkak256("SubAddr" + master_priv_vkey + major_idx + minor_idx)
        m = Kekkak256.QuickDigest(MoneroSubaddressConst.SUBADDR_PREFIX
                                  + self.m_priv_vkey.Raw().ToBytes()
                                  + major_idx_bytes
                                  + minor_idx_bytes)
        m_int = Ed25519Utils.IntDecode(Ed25519Utils.ScalarReduce(m))

        # Compute subaddress public spend key
        # D = master_pub_skey + m * B
        subaddr_pub_skey_point = self.m_pub_skey.KeyObject().Point() + (Ed25519Monero.Generator() * m_int)

        # Compute subaddress public view key
        # C = master_priv_vkey * D
        subaddr_pub_vkey_point = subaddr_pub_skey_point * self.m_priv_vkey.Raw().ToInt("little")

        return (MoneroPublicKey.FromPoint(subaddr_pub_skey_point),
                MoneroPublicKey.FromPoint(subaddr_pub_vkey_point))

    def ComputeAndEncodeKeys(self,
                             minor_idx: int,
                             major_idx: int,
                             net_ver: bytes) -> str:
        """
        Compute the public keys of the specified subaddress and encode them.

        Args:
            minor_idx (int): Minor index (i.e. subaddress index)
            major_idx (int): Major index (i.e. account index)
            net_ver (bytes): Net version

        Returns:
            str: Encoded subaddress string

        Raises:
            ValueError: If one of the indexes is not valid
        """
        pub_skey, pub_vkey = self.ComputeKeys(minor_idx, major_idx)

        # Encode subaddress
        return XmrAddrEncoder.EncodeKey(pub_skey.KeyObject(),
                                        pub_vkey=pub_vkey.KeyObject(),
                                        net_ver=net_ver)



from bip_utils.electrum.mnemonic_v1.electrum_v1_entropy_generator import (
    ElectrumV1EntropyBitLen, ElectrumV1EntropyGenerator
)
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic import (
    ElectrumV1Languages, ElectrumV1Mnemonic, ElectrumV1WordsNum
)
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_decoder import ElectrumV1MnemonicDecoder
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_encoder import ElectrumV1MnemonicEncoder
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_generator import ElectrumV1MnemonicGenerator
from bip_utils.electrum.mnemonic_v1.electrum_v1_mnemonic_validator import ElectrumV1MnemonicValidator
from bip_utils.electrum.mnemonic_v1.electrum_v1_seed_generator import ElectrumV1SeedGenerator



from bip_utils.addr.ada_byron_addr import (
    AdaByronAddrDecoder, AdaByronAddrTypes, AdaByronIcarusAddr, AdaByronIcarusAddrEncoder, AdaByronLegacyAddr,
    AdaByronLegacyAddrEncoder
)
from bip_utils.addr.ada_shelley_addr import (
    AdaShelleyAddr, AdaShelleyAddrDecoder, AdaShelleyAddrEncoder, AdaShelleyAddrNetworkTags, AdaShelleyRewardAddr,
    AdaShelleyRewardAddrDecoder, AdaShelleyRewardAddrEncoder, AdaShelleyStakingAddr, AdaShelleyStakingAddrDecoder,
    AdaShelleyStakingAddrEncoder
)
from bip_utils.addr.algo_addr import AlgoAddr, AlgoAddrDecoder, AlgoAddrEncoder
from bip_utils.addr.aptos_addr import AptosAddr, AptosAddrDecoder, AptosAddrEncoder
from bip_utils.addr.atom_addr import AtomAddr, AtomAddrDecoder, AtomAddrEncoder
from bip_utils.addr.avax_addr import (
    AvaxPChainAddr, AvaxPChainAddrDecoder, AvaxPChainAddrEncoder, AvaxXChainAddr, AvaxXChainAddrDecoder,
    AvaxXChainAddrEncoder
)
from bip_utils.addr.bch_addr_converter import BchAddrConverter
from bip_utils.addr.egld_addr import EgldAddr, EgldAddrDecoder, EgldAddrEncoder
from bip_utils.addr.eos_addr import EosAddr, EosAddrDecoder, EosAddrEncoder
from bip_utils.addr.ergo_addr import ErgoNetworkTypes, ErgoP2PKHAddr, ErgoP2PKHAddrDecoder, ErgoP2PKHAddrEncoder
from bip_utils.addr.eth_addr import EthAddr, EthAddrDecoder, EthAddrEncoder
from bip_utils.addr.fil_addr import FilSecp256k1Addr, FilSecp256k1AddrDecoder, FilSecp256k1AddrEncoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.addr.icx_addr import IcxAddr, IcxAddrDecoder, IcxAddrEncoder
from bip_utils.addr.nano_addr import NanoAddr, NanoAddrDecoder, NanoAddrEncoder
from bip_utils.addr.near_addr import NearAddr, NearAddrDecoder, NearAddrEncoder
from bip_utils.addr.neo_addr import NeoAddr, NeoAddrDecoder, NeoAddrEncoder
from bip_utils.addr.okex_addr import OkexAddr, OkexAddrDecoder, OkexAddrEncoder
from bip_utils.addr.one_addr import OneAddr, OneAddrDecoder, OneAddrEncoder
from bip_utils.addr.P2PKH_addr import (
    BchP2PKHAddr, BchP2PKHAddrDecoder, BchP2PKHAddrEncoder, P2PKHAddr, P2PKHAddrDecoder, P2PKHAddrEncoder,
    P2PKHPubKeyModes
)
from bip_utils.addr.P2SH_addr import (
    BchP2SHAddr, BchP2SHAddrDecoder, BchP2SHAddrEncoder, P2SHAddr, P2SHAddrDecoder, P2SHAddrEncoder
)
from bip_utils.addr.P2TR_addr import P2TRAddr, P2TRAddrDecoder, P2TRAddrEncoder
from bip_utils.addr.P2WPKH_addr import P2WPKHAddr, P2WPKHAddrDecoder, P2WPKHAddrEncoder
from bip_utils.addr.sol_addr import SolAddr, SolAddrDecoder, SolAddrEncoder
from bip_utils.addr.substrate_addr import (
    SubstrateEd25519Addr, SubstrateEd25519AddrDecoder, SubstrateEd25519AddrEncoder, SubstrateSr25519Addr,
    SubstrateSr25519AddrDecoder, SubstrateSr25519AddrEncoder
)
from bip_utils.addr.trx_addr import TrxAddr, TrxAddrDecoder, TrxAddrEncoder
from bip_utils.addr.xlm_addr import XlmAddr, XlmAddrDecoder, XlmAddrEncoder, XlmAddrTypes
from bip_utils.addr.xmr_addr import (
    XmrAddr, XmrAddrDecoder, XmrAddrEncoder, XmrIntegratedAddr, XmrIntegratedAddrDecoder, XmrIntegratedAddrEncoder
)
from bip_utils.addr.xrp_addr import XrpAddr, XrpAddrDecoder, XrpAddrEncoder
from bip_utils.addr.xtz_addr import XtzAddr, XtzAddrDecoder, XtzAddrEncoder, XtzAddrPrefixes
from bip_utils.addr.zil_addr import ZilAddr, ZilAddrDecoder, ZilAddrEncoder



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Neo address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.base58 import Base58ChecksumError, Base58Decoder, Base58Encoder
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import Hash160
from bip_utils.utils.misc import BytesUtils, IntegerUtils


class NeoAddrConst:
    """Class container for NEO address constants."""

    # Address prefix byte
    PREFIX_BYTE: bytes = b"\x21"
    # Address suffix byte
    SUFFIX_BYTE: bytes = b"\xac"


class NeoAddrDecoder(IAddrDecoder):
    """
    Neo address decoder class.
    It allows the Neo address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Neo address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            ver (bytes): Expected version

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """
        ver_bytes = kwargs["ver"]

        try:
            # Decode from base58
            addr_dec_bytes = Base58Decoder.CheckDecode(addr)
        except Base58ChecksumError as ex:
            raise ValueError("Invalid base58 checksum") from ex

        # Validate length
        AddrDecUtils.ValidateLength(addr_dec_bytes,
                                    Hash160.DigestSize() + len(ver_bytes))
        # Check version
        ver_got = IntegerUtils.ToBytes(addr_dec_bytes[0])
        if ver_bytes != ver_got:
            raise ValueError(f"Invalid version (expected {BytesUtils.ToHexString(ver_bytes)}, "
                             f"got {BytesUtils.ToHexString(ver_got)})")

        return addr_dec_bytes[1:]


class NeoAddrEncoder(IAddrEncoder):
    """
    Neo address encoder class.
    It allows the Neo address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Neo address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object

        Other Parameters:
            ver (bytes): Version

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519
        """
        ver_bytes = kwargs["ver"]

        pub_key_obj = AddrKeyValidator.ValidateAndGetNist256p1Key(pub_key)

        # Get payload
        payload_bytes = (NeoAddrConst.PREFIX_BYTE
                         + pub_key_obj.RawCompressed().ToBytes()
                         + NeoAddrConst.SUFFIX_BYTE)
        # Encode to base58
        return Base58Encoder.CheckEncode(ver_bytes + Hash160.QuickDigest(payload_bytes))


# Deprecated: only for compatibility, Encoder class shall be used instead
NeoAddr = NeoAddrEncoder



from bip_utils.bip.bip32.base import Bip32Base, IBip32KeyDerivator, IBip32MstKeyGenerator
from bip_utils.bip.bip32.bip32_const import Bip32Const
from bip_utils.bip.bip32.bip32_ex import Bip32KeyError, Bip32PathError
from bip_utils.bip.bip32.bip32_key_data import Bip32ChainCode, Bip32Depth, Bip32FingerPrint, Bip32KeyData, Bip32KeyIndex
from bip_utils.bip.bip32.bip32_key_net_ver import Bip32KeyNetVersions
from bip_utils.bip.bip32.bip32_key_ser import (
    Bip32DeserializedKey, Bip32KeyDeserializer, Bip32PrivateKeySerializer, Bip32PublicKeySerializer
)
from bip_utils.bip.bip32.bip32_keys import Bip32PrivateKey, Bip32PublicKey
from bip_utils.bip.bip32.bip32_path import Bip32Path, Bip32PathParser
from bip_utils.bip.bip32.bip32_utils import Bip32Utils
from bip_utils.bip.bip32.kholaw import (
    Bip32Ed25519Kholaw, Bip32KholawEd25519, Bip32KholawEd25519KeyDerivator, Bip32KholawEd25519KeyDerivatorBase,
    Bip32KholawEd25519MstKeyGenerator
)
from bip_utils.bip.bip32.slip10 import (
    Bip32Ed25519Blake2bSlip, Bip32Ed25519Slip, Bip32Nist256p1, Bip32Secp256k1, Bip32Slip10EcdsaDerivator,
    Bip32Slip10Ed2519MstKeyGenerator, Bip32Slip10Ed25519, Bip32Slip10Ed25519Blake2b, Bip32Slip10Ed25519Derivator,
    Bip32Slip10Nist256p1, Bip32Slip10Nist256p1MstKeyGenerator, Bip32Slip10Secp256k1, Bip32Slip10Secp256k1MstKeyGenerator
)



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module containing utility classes for Electrum v2 keys derivation, since it uses its own paths."""

# Imports
from __future__ import annotations

from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Union

from bip_utils.addr import P2PKHAddr, P2WPKHAddr
from bip_utils.bip.bip32 import Bip32Base, Bip32KeyIndex, Bip32PrivateKey, Bip32PublicKey, Bip32Slip10Secp256k1
from bip_utils.coin_conf import CoinsConf


class ElectrumV2Base(ABC):
    """Electrum v2 base class."""

    m_bip32_obj: Bip32Base

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes) -> ElectrumV2Base:
        """
        Construct class from seed bytes.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            ElectrumV2Base object: ElectrumV2Base object
        """
        return cls(Bip32Slip10Secp256k1.FromSeed(seed_bytes))

    def __init__(self,
                 bip32_obj: Bip32Base) -> None:
        """
        Construct class.

        Args:
            bip32_obj (Bip32Base object): Bip32Base object (shall be a Bip32Slip10Secp256k1 instance)

        Raises:
            TypeError: If the bip32 object is not a Bip32Slip10Secp256k1 class instance
            ValueError: If the bip32 object is not a master key
        """
        if not isinstance(bip32_obj, Bip32Slip10Secp256k1):
            raise TypeError("A Bip32Slip10Secp256k1 class instance is required")
        if bip32_obj.Depth() > 0:
            raise ValueError("The Bip32 object shall be a master key (i.e. depth equal to 0)")
        self.m_bip32_obj = bip32_obj

    def Bip32Object(self) -> Bip32Base:
        """
        Return the BIP32 object.

        Returns:
            Bip32Base object: Bip32Base object
        """
        return self.m_bip32_obj

    def IsPublicOnly(self) -> bool:
        """
        Get if it's public-only.

        Returns:
            bool: True if public-only, false otherwise
        """
        return self.m_bip32_obj.IsPublicOnly()

    def MasterPrivateKey(self) -> Bip32PrivateKey:
        """
        Get the master private key.

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object
        """
        return self.m_bip32_obj.PrivateKey()

    def MasterPublicKey(self) -> Bip32PublicKey:
        """
        Get the master public key.

        Returns:
            Bip32PublicKey object: Bip32PublicKey object
        """
        return self.m_bip32_obj.PublicKey()

    @abstractmethod
    def GetPrivateKey(self,
                      change_idx: Union[int, Bip32KeyIndex],
                      addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PrivateKey:
        """
        Get the private key with the specified change and address indexes.

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key or the object is public-only
            Bip32PathError: If the path indexes are not valid
        """

    @abstractmethod
    def GetPublicKey(self,
                     change_idx: Union[int, Bip32KeyIndex],
                     addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PublicKey:
        """
        Get the public key with the specified change and address indexes.

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PublicKey object: Bip32PublicKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """

    @abstractmethod
    def GetAddress(self,
                   change_idx: Union[int, Bip32KeyIndex],
                   addr_idx: Union[int, Bip32KeyIndex]) -> str:
        """
        Get the address with the specified change and address indexes.

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            str: Address

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """


class ElectrumV2Standard(ElectrumV2Base):
    """
    Electrum v2 standard class.
    It derives keys like the Electrum wallet with standard mnemonic.
    """

    def GetPrivateKey(self,
                      change_idx: Union[int, Bip32KeyIndex],
                      addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PrivateKey:
        """
        Get the private key with the specified change and address indexes.
        Derivation path: m/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key or the object is public-only
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(change_idx, addr_idx).PrivateKey()

    def GetPublicKey(self,
                     change_idx: Union[int, Bip32KeyIndex],
                     addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PublicKey:
        """
        Get the public key with the specified change and address indexes.
        Derivation path: m/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PublicKey object: Bip32PublicKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(change_idx, addr_idx).PublicKey()

    @lru_cache()
    def GetAddress(self,
                   change_idx: Union[int, Bip32KeyIndex],
                   addr_idx: Union[int, Bip32KeyIndex]) -> str:
        """
        Get the address with the specified change and address indexes.
        Derivation path: m/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            str: Address

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return P2PKHAddr.EncodeKey(self.GetPublicKey(change_idx, addr_idx).KeyObject(),
                                   net_ver=CoinsConf.BitcoinMainNet.ParamByKey("p2pkh_net_ver"))

    @lru_cache()
    def __DeriveKey(self,
                    change_idx: Union[int, Bip32KeyIndex],
                    addr_idx: Union[int, Bip32KeyIndex]) -> Bip32Base:
        """
        Derive the key with the specified change and address indexes.
        Derivation path: m/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32Base object: Bip32Base object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.m_bip32_obj.DerivePath(f"m/{change_idx}/{addr_idx}")


class ElectrumV2Segwit(ElectrumV2Base):
    """
    Electrum v2 segwit class.
    It derives keys like the Electrum wallet with segwit mnemonic.
    """

    m_bip32_acc: Bip32Base

    def __init__(self,
                 bip32: Bip32Base) -> None:
        """
        Construct class.

        Args:
            bip32 (Bip32Base object): Bip32Base object
        """
        super().__init__(bip32)
        self.m_bip32_acc = bip32.DerivePath("m/0'")

    def GetPrivateKey(self,
                      change_idx: Union[int, Bip32KeyIndex],
                      addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PrivateKey:
        """
        Get the private key with the specified change and address indexes.
        Derivation path: m/0'/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key or the object is public-only
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(change_idx, addr_idx).PrivateKey()

    def GetPublicKey(self,
                     change_idx: Union[int, Bip32KeyIndex],
                     addr_idx: Union[int, Bip32KeyIndex]) -> Bip32PublicKey:
        """
        Get the public key with the specified change and address indexes.
        Derivation path: m/0'/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32PublicKey object: Bip32PublicKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(change_idx, addr_idx).PublicKey()

    @lru_cache()
    def GetAddress(self,
                   change_idx: Union[int, Bip32KeyIndex],
                   addr_idx: Union[int, Bip32KeyIndex]) -> str:
        """
        Get the address with the specified change and address indexes.
        Derivation path: m/0'/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            str: Address

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return P2WPKHAddr.EncodeKey(self.GetPublicKey(change_idx, addr_idx).KeyObject(),
                                    hrp=CoinsConf.BitcoinMainNet.ParamByKey("p2wpkh_hrp"))

    @lru_cache()
    def __DeriveKey(self,
                    change_idx: Union[int, Bip32KeyIndex],
                    addr_idx: Union[int, Bip32KeyIndex]) -> Bip32Base:
        """
        Derive the key with the specified change and address indexes.
        Derivation path: m/0'/change_idx/addr_idx

        Args:
            change_idx (int or Bip32KeyIndex object): Change index
            addr_idx (int or Bip32KeyIndex object)  : Address index

        Returns:
            Bip32Base object: Bip32Base object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.m_bip32_acc.DerivePath(f"{change_idx}/{addr_idx}")



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Cardano Byron legacy keys derivation."""


# Imports
from __future__ import annotations

from functools import lru_cache
from typing import Union

from bip_utils.addr import AdaByronAddrDecoder, AdaByronLegacyAddrEncoder
from bip_utils.bip.bip32 import Bip32Base, Bip32KeyIndex, Bip32Path, Bip32PrivateKey, Bip32PublicKey
from bip_utils.cardano.bip32 import CardanoByronLegacyBip32
from bip_utils.utils.crypto import Pbkdf2HmacSha512


class CardanoByronLegacyConst:
    """Class container for Cardano Byron legacy constants."""

    # PBKDF2 salt used for deriving the HD path key
    HD_PATH_KEY_PBKDF2_SALT: str = "address-hashing"
    # PBKDF2 rounds used for deriving the HD path key
    HD_PATH_KEY_PBKDF2_ROUNDS: int = 500
    # PBKDF2 output byte length used for deriving the HD path key
    HD_PATH_KEY_PBKDF2_OUT_BYTE_LEN: int = 32


class CardanoByronLegacy:
    """
    Cardano Byron legacy class.
    It allows master key generation, children keys derivation and addresses computation like the old Daedalus wallet.
    Addresses are in the Ddz... format, which contains the encrypted derivation path.
    """

    m_bip32_obj: Bip32Base

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes) -> CardanoByronLegacy:
        """
        Construct class from seed bytes.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            CardanoByronLegacy object: CardanoByronLegacy object
        """
        return cls(CardanoByronLegacyBip32.FromSeed(seed_bytes))

    def __init__(self,
                 bip32_obj: Bip32Base) -> None:
        """
        Construct class.

        Args:
            bip32_obj (Bip32Base object): Bip32Base object

        Raises:
            ValueError: If the bip32 object is not a master key
            TypeError: If the bip32 object is not a CardanoByronLegacyBip32 class instance
        """
        if not isinstance(bip32_obj, CardanoByronLegacyBip32):
            raise TypeError("The Bip32 object shall be a CardanoByronLegacyBip32 instance")
        if bip32_obj.Depth() != 0:
            raise ValueError("The Bip32 object shall be a master key (i.e. depth equal to 0)")
        self.m_bip32_obj = bip32_obj

    def Bip32Object(self) -> Bip32Base:
        """
        Return the BIP32 object.

        Returns:
            Bip32Base object: Bip32Base object
        """
        return self.m_bip32_obj

    @lru_cache()
    def HdPathKey(self) -> bytes:
        """
        Get the key used for HD path decryption/encryption.

        Returns:
            bytes: Key bytes
        """
        return Pbkdf2HmacSha512.DeriveKey(
            self.m_bip32_obj.PublicKey().RawCompressed().ToBytes()[1:] + self.m_bip32_obj.ChainCode().ToBytes(),
            CardanoByronLegacyConst.HD_PATH_KEY_PBKDF2_SALT,
            CardanoByronLegacyConst.HD_PATH_KEY_PBKDF2_ROUNDS,
            CardanoByronLegacyConst.HD_PATH_KEY_PBKDF2_OUT_BYTE_LEN
        )

    def HdPathFromAddress(self,
                          address: str) -> Bip32Path:
        """
        Get the HD path from an address by decrypting it.
        The address shall be derived from the current object master key (i.e. self.m_bip32_obj) in order to
        successfully decrypt the path.

        Args:
            address (str): Address string

        Returns:
            Bip32Path object: Bip32Path object

        Raises:
            ValueError: If the address encoding is not valid or the path cannot be decrypted
        """
        addr_dec_bytes = AdaByronAddrDecoder.DecodeAddr(address)
        hd_path_dec_bytes = AdaByronAddrDecoder.DecryptHdPath(AdaByronAddrDecoder.SplitDecodedBytes(addr_dec_bytes)[1],
                                                              self.HdPathKey())
        return hd_path_dec_bytes

    def MasterPrivateKey(self) -> Bip32PrivateKey:
        """
        Get the master private key.

        Returns:
            Bip32PrivateKey object: Bip32PrivateKey object
        """
        return self.m_bip32_obj.PrivateKey()

    def MasterPublicKey(self) -> Bip32PublicKey:
        """
        Get the master public key.

        Returns:
            Bip32PublicKey object: Bip32PublicKey object
        """
        return self.m_bip32_obj.PublicKey()

    def GetPrivateKey(self,
                      first_idx: Union[int, Bip32KeyIndex],
                      second_idx: Union[int, Bip32KeyIndex]) -> Bip32PrivateKey:
        """
        Get the private key with the specified indexes.
        Derivation path: m/first_idx'/second_idx'
        The indexes will be automatically hardened if not (e.g. 0, 1' -> 0', 1').

        Args:
            first_idx (int or Bip32KeyIndex object) : First index
            second_idx (int or Bip32KeyIndex object): Second index

        Returns:
            IPrivateKey object: IPrivateKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(first_idx, second_idx).PrivateKey()

    def GetPublicKey(self,
                     first_idx: Union[int, Bip32KeyIndex],
                     second_idx: Union[int, Bip32KeyIndex]) -> Bip32PublicKey:
        """
        Get the public key with the specified indexes.
        Derivation path: m/first_idx'/second_idx'
        The indexes will be automatically hardened if not (e.g. 0, 1' -> 0', 1').


        Args:
            first_idx (int or Bip32KeyIndex object) : First index
            second_idx (int or Bip32KeyIndex object): Second index

        Returns:
            IPublicKey object: IPublicKey object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.__DeriveKey(first_idx, second_idx).PublicKey()

    @lru_cache()
    def GetAddress(self,
                   first_idx: Union[int, Bip32KeyIndex],
                   second_idx: Union[int, Bip32KeyIndex]) -> str:
        """
        Get the address with the specified indexes.
        Derivation path: m/first_idx'/second_idx'
        The indexes will be automatically hardened if not (e.g. 0, 1' -> 0', 1').

        Args:
            first_idx (int or Bip32KeyIndex object) : First index
            second_idx (int or Bip32KeyIndex object): Second index

        Returns:
            str: Address

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        pub_key = self.GetPublicKey(first_idx, second_idx)
        return AdaByronLegacyAddrEncoder.EncodeKey(
            pub_key.KeyObject(),
            chain_code=pub_key.ChainCode(),
            hd_path=self.__GetDerivationPath(first_idx, second_idx),
            hd_path_key=self.HdPathKey()
        )

    @lru_cache()
    def __DeriveKey(self,
                    first_idx: Union[int, Bip32KeyIndex],
                    second_idx: Union[int, Bip32KeyIndex]) -> Bip32Base:
        """
        Derive the key with the specified change and address indexes.
        Derivation path: m/first_idx'/second_idx'

        Args:
            first_idx (int or Bip32KeyIndex object) : First index
            second_idx (int or Bip32KeyIndex object): Second index

        Returns:
            Bip32Base object: Bip32Base object

        Raises:
            Bip32KeyError: If the derivation results in an invalid key
            Bip32PathError: If the path indexes are not valid
        """
        return self.m_bip32_obj.DerivePath(
            self.__GetDerivationPath(first_idx, second_idx)
        )

    @staticmethod
    def __GetDerivationPath(first_idx: Union[int, Bip32KeyIndex],
                            second_idx: Union[int, Bip32KeyIndex]) -> str:
        """
        Get the derivation path from the specified indexes.

        Args:
            first_idx (int or Bip32KeyIndex object) : First index
            second_idx (int or Bip32KeyIndex object): Second index

        Returns:
            str: Derivation path
        """
        return f"m/{first_idx}'/{second_idx}'"



from bip_utils.bip.bip39.bip39_entropy_generator import Bip39EntropyBitLen, Bip39EntropyGenerator
from bip_utils.bip.bip39.bip39_mnemonic import Bip39Languages, Bip39Mnemonic, Bip39WordsNum
from bip_utils.bip.bip39.bip39_mnemonic_decoder import Bip39MnemonicDecoder
from bip_utils.bip.bip39.bip39_mnemonic_encoder import Bip39MnemonicEncoder
from bip_utils.bip.bip39.bip39_mnemonic_generator import Bip39MnemonicGenerator
from bip_utils.bip.bip39.bip39_mnemonic_validator import Bip39MnemonicValidator
from bip_utils.bip.bip39.bip39_seed_generator import Bip39SeedGenerator
from bip_utils.bip.bip39.ibip39_seed_generator import IBip39SeedGenerator



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Algorand mnemonic."""

# Imports
from enum import IntEnum, unique
from typing import List

from bip_utils.bip.bip39 import Bip39Languages, Bip39Mnemonic
from bip_utils.utils.mnemonic import MnemonicLanguages


@unique
class AlgorandWordsNum(IntEnum):
    """Enumerative for Algorand words number."""

    WORDS_NUM_25 = 25


@unique
class AlgorandLanguages(MnemonicLanguages):
    """Enumerative for Algorand languages."""

    ENGLISH = Bip39Languages.ENGLISH


class AlgorandMnemonicConst:
    """Class container for Algorand mnemonic constants."""

    # Accepted mnemonic word numbers
    MNEMONIC_WORD_NUM: List[AlgorandWordsNum] = [
        AlgorandWordsNum.WORDS_NUM_25,
    ]

    # Checksum length in bytes
    CHECKSUM_BYTE_LEN: int = 2


class AlgorandMnemonic(Bip39Mnemonic):
    """Algorand mnemonic class."""



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Ethereum address encoding/decoding."""

# Imports
from typing import Any, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.coin_conf import CoinsConf
from bip_utils.ecc import IPublicKey
from bip_utils.utils.crypto import Kekkak256
from bip_utils.utils.misc import BytesUtils


class EthAddrConst:
    """Class container for Ethereum address constants."""

    # Start byte
    START_BYTE: int = 24
    # Address length
    ADDR_LEN: int = 40


class _EthAddrUtils:
    """Class container for Ethereum address utility functions."""

    @staticmethod
    def ChecksumEncode(addr: str) -> str:
        """
        Checksum encode the specified address.

        Args:
            addr (str): Address string

        Returns:
            str: Checksum encoded address
        """

        # Compute address digest
        addr_hex_digest = BytesUtils.ToHexString(Kekkak256.QuickDigest(addr.lower()))
        # Encode it
        enc_addr = [c.upper() if (int(addr_hex_digest[i], 16) >= 8) else c.lower() for i, c in enumerate(addr)]

        return "".join(enc_addr)


class EthAddrDecoder(IAddrDecoder):
    """
    Ethereum address decoder class.
    It allows the Ethereum address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode an Ethereum address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            skip_chksum_enc (bool, optional): True to skip checksum encoding verification, false otherwise (default)

        Returns:
            bytes: Public key hash bytes

        Raises:
            ValueError: If the address encoding is not valid
        """
        skip_chksum_enc = kwargs.get("skip_chksum_enc", False)

        # Validate and remove prefix
        addr_no_prefix = AddrDecUtils.ValidateAndRemovePrefix(addr,
                                                              CoinsConf.Ethereum.ParamByKey("addr_prefix"))
        # Validate length
        AddrDecUtils.ValidateLength(addr_no_prefix, EthAddrConst.ADDR_LEN)
        # Check checksum encoding
        if not skip_chksum_enc and addr_no_prefix != _EthAddrUtils.ChecksumEncode(addr_no_prefix):
            raise ValueError("Invalid checksum encoding")

        return BytesUtils.FromHexString(addr_no_prefix)


class EthAddrEncoder(IAddrEncoder):
    """
    Ethereum address encoder class.
    It allows the Ethereum address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Ethereum address.

        Args:
            pub_key (bytes or IPublicKey): Public key bytes or object

        Other Parameters:
            skip_chksum_enc (bool, optional): True to skip checksum encoding, false otherwise (default)

        Returns:
            str: Address string

        Raised:
            ValueError: If the public key is not valid
            TypeError: If the public key is not secp256k1
        """
        skip_chksum_enc = kwargs.get("skip_chksum_enc", False)

        pub_key_obj = AddrKeyValidator.ValidateAndGetSecp256k1Key(pub_key)

        # First byte of the uncompressed key (i.e. 0x04) is not needed
        kekkak_hex = BytesUtils.ToHexString(Kekkak256.QuickDigest(pub_key_obj.RawUncompressed().ToBytes()[1:]))
        addr = kekkak_hex[EthAddrConst.START_BYTE:]
        return CoinsConf.Ethereum.ParamByKey("addr_prefix") + (_EthAddrUtils.ChecksumEncode(addr)
                                                               if not skip_chksum_enc
                                                               else addr)


# Deprecated: only for compatibility, Encoder class shall be used instead
EthAddr = EthAddrEncoder



from bip_utils.monero.conf.monero_coin_conf import MoneroCoinConf
from bip_utils.monero.conf.monero_coins import MoneroCoins
from bip_utils.monero.conf.monero_conf import MoneroConf
from bip_utils.monero.conf.monero_conf_getter import MoneroConfGetter



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Algorand mnemonic encoding.
Reference: https://github.com/algorand/py-algorand-sdk
"""

# Imports
from typing import List

from bip_utils.algorand.mnemonic.algorand_entropy_generator import AlgorandEntropyGenerator
from bip_utils.algorand.mnemonic.algorand_mnemonic import AlgorandLanguages, AlgorandMnemonic
from bip_utils.algorand.mnemonic.algorand_mnemonic_utils import AlgorandMnemonicUtils
from bip_utils.bip.bip39.bip39_mnemonic_utils import Bip39WordsListGetter
from bip_utils.utils.mnemonic import Mnemonic, MnemonicEncoderBase


class AlgorandMnemonicEncoder(MnemonicEncoderBase):
    """
    Algorand mnemonic encoder class.
    It encodes bytes to the mnemonic phrase.
    """

    def __init__(self,
                 lang: AlgorandLanguages = AlgorandLanguages.ENGLISH) -> None:
        """
        Construct class.

        Args:
            lang (AlgorandLanguages, optional): Language (default: English)

        Raises:
            TypeError: If the language is not a AlgorandLanguages enum
            ValueError: If loaded words list is not valid
        """
        if not isinstance(lang, AlgorandLanguages):
            raise TypeError("Language is not an enumerative of AlgorandLanguages")
        super().__init__(lang.value, Bip39WordsListGetter)

    def Encode(self,
               entropy_bytes: bytes) -> Mnemonic:
        """
        Encode bytes to mnemonic phrase.

        Args:
            entropy_bytes (bytes): Entropy bytes

        Returns:
            Mnemonic object: Encoded mnemonic

        Raises:
            ValueError: If bytes length is not valid
        """

        # Check entropy length
        entropy_byte_len = len(entropy_bytes)
        if not AlgorandEntropyGenerator.IsValidEntropyByteLen(entropy_byte_len):
            raise ValueError(f"Entropy byte length ({entropy_byte_len}) is not valid")

        # Compute checksum word
        chksum_word_idx = AlgorandMnemonicUtils.ComputeChecksumWordIndex(entropy_bytes)
        # Convert entropy bytes to a list of word indexes
        word_indexes = AlgorandMnemonicUtils.ConvertBits(entropy_bytes, 8, 11)
        # Cannot be None by converting bytes from 8-bit to 11-bit
        assert word_indexes is not None
        # Get mnemonic
        return AlgorandMnemonic.FromList(self.__IndexesToWords(word_indexes + [chksum_word_idx]))

    def __IndexesToWords(self,
                         indexes: List[int]) -> List[str]:
        """
        Get a list of words from a list of indexes.

        Args:
            indexes (list[int]): List of indexes

        Returns:
            list[str]: List of words
        """
        return [self.m_words_list.GetWordAtIdx(idx) for idx in indexes]



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Monero keys computation and derivation."""

# Imports
from __future__ import annotations

from functools import lru_cache
from typing import Optional, Union

from bip_utils.addr import XmrIntegratedAddrEncoder
from bip_utils.ecc import Ed25519MoneroPrivateKey, Ed25519Utils, IPrivateKey, IPublicKey
from bip_utils.monero.conf import MoneroCoinConf, MoneroCoins, MoneroConfGetter
from bip_utils.monero.monero_ex import MoneroKeyError
from bip_utils.monero.monero_keys import MoneroPrivateKey, MoneroPublicKey
from bip_utils.monero.monero_subaddr import MoneroSubaddress
from bip_utils.utils.crypto import Kekkak256


class Monero:
    """
    Monero class.
    It allows to compute Monero keys and addresses/subaddresses.
    """

    m_priv_skey: Optional[MoneroPrivateKey]
    m_priv_vkey: MoneroPrivateKey
    m_pub_skey: MoneroPublicKey
    m_pub_vkey: MoneroPublicKey
    m_coin_conf: MoneroCoinConf
    m_subaddr: MoneroSubaddress

    @classmethod
    def FromSeed(cls,
                 seed_bytes: bytes,
                 coin_type: MoneroCoins = MoneroCoins.MONERO_MAINNET) -> Monero:
        """
        Create from seed bytes.

        Args:
            seed_bytes (bytes)               : Seed bytes
            coin_type (MoneroCoins, optional): Coin type (default: main net)

        Returns:
            Monero object: Monero object
        """
        priv_skey_bytes = (seed_bytes
                           if len(seed_bytes) == Ed25519MoneroPrivateKey.Length()
                           else Kekkak256.QuickDigest(seed_bytes))
        return cls.FromPrivateSpendKey(
            Ed25519Utils.ScalarReduce(priv_skey_bytes),
            coin_type
        )

    @classmethod
    def FromBip44PrivateKey(cls,
                            priv_key: Union[bytes, IPrivateKey],
                            coin_type: MoneroCoins = MoneroCoins.MONERO_MAINNET) -> Monero:
        """
        Create from Bip44 private key bytes.

        Args:
            priv_key (bytes or IPrivateKey)  : Private key
            coin_type (MoneroCoins, optional): Coin type (default: main net)

        Returns:
            Monero object: Monero object
        """
        if not isinstance(priv_key, bytes):
            priv_key = priv_key.Raw().ToBytes()
        return cls.FromPrivateSpendKey(
            Ed25519Utils.ScalarReduce(Kekkak256.QuickDigest(priv_key)),
            coin_type
        )

    @classmethod
    def FromPrivateSpendKey(cls,
                            priv_skey: Union[bytes, IPrivateKey],
                            coin_type: MoneroCoins = MoneroCoins.MONERO_MAINNET) -> Monero:
        """
        Create from private spend key.

        Args:
            priv_skey (bytes or IPrivateKey) : Private spend key
            coin_type (MoneroCoins, optional): Coin type (default: main net)

        Returns:
            Monero object: Monero object

        Raises:
            MoneroKeyError: If the key constructed from the bytes is not valid
        """
        return cls(priv_key=priv_skey,
                   coin_type=coin_type)

    @classmethod
    def FromWatchOnly(cls,
                      priv_vkey: Union[bytes, IPrivateKey],
                      pub_skey: Union[bytes, IPublicKey],
                      coin_type: MoneroCoins = MoneroCoins.MONERO_MAINNET) -> Monero:
        """
        Create from private view key and public spend key (i.e. watch-only wallet).

        Args:
            priv_vkey (bytes or IPrivateKey) : Private view key
            pub_skey (bytes or IPublicKey)   : Public spend key
            coin_type (MoneroCoins, optional): Coin type (default: main net)

        Returns:
            Monero object: Monero object

        Raises:
            MoneroKeyError: If the key constructed from the bytes is not valid
        """
        return cls(priv_key=priv_vkey,
                   pub_key=pub_skey,
                   coin_type=coin_type)

    def __init__(self,
                 priv_key: Union[bytes, IPrivateKey],
                 pub_key: Optional[Union[bytes, IPublicKey]] = None,
                 coin_type: MoneroCoins = MoneroCoins.MONERO_MAINNET) -> None:
        """
        Construct class.

        Args:
            priv_key (bytes or IPrivateKey)  : Private key (view key if watch-only wallet, otherwise spend key)
            pub_key (bytes or IPublicKey)    : Public spend key (only needed for watch-only wallets, otherwise None)
            coin_type (MoneroCoins, optional): Coin type (default: main net)

        Raises:
            MoneroKeyError: If the key constructed from the bytes is not valid
        """

        # Private key object
        if pub_key is None:
            self.m_priv_skey = MoneroPrivateKey.FromBytesOrKeyObject(priv_key)
            self.m_priv_vkey = self.__ViewFromSpendKey(self.m_priv_skey)
            self.m_pub_skey = self.m_priv_skey.PublicKey()
            self.m_pub_vkey = self.m_priv_vkey.PublicKey()
        # Watch-only object
        else:
            self.m_priv_skey = None
            self.m_priv_vkey = MoneroPrivateKey.FromBytesOrKeyObject(priv_key)
            self.m_pub_skey = MoneroPublicKey.FromBytesOrKeyObject(pub_key)
            self.m_pub_vkey = self.m_priv_vkey.PublicKey()

        self.m_coin_conf = MoneroConfGetter.GetConfig(coin_type)
        self.m_subaddr = MoneroSubaddress(self.m_priv_vkey, self.m_pub_skey, self.m_pub_vkey)

    def IsWatchOnly(self) -> bool:
        """
        Return if it's a watch-only object.

        Returns:
            bool: True if watch-only, false otherwise
        """
        return self.m_priv_skey is None

    def CoinConf(self) -> MoneroCoinConf:
        """
        Return coin configuration.

        Returns:
            MoneroCoinConf object: MoneroCoinConf object
        """
        return self.m_coin_conf

    def PrivateSpendKey(self) -> MoneroPrivateKey:
        """
        Return the private spend key.

        Returns:
            MoneroPrivateKey object: MoneroPrivateKey object

        Raises:
            MoneroKeyError: If the class is watch-only
        """
        if self.IsWatchOnly():
            raise MoneroKeyError("Watch-only class has not a private spend key")

        assert isinstance(self.m_priv_skey, MoneroPrivateKey)
        return self.m_priv_skey

    def PrivateViewKey(self) -> MoneroPrivateKey:
        """
        Return the private view key.

        Returns:
            MoneroPrivateKey object: MoneroPrivateKey object
        """
        return self.m_priv_vkey

    def PublicSpendKey(self) -> MoneroPublicKey:
        """
        Return the public spend key.

        Returns:
            MoneroPublicKey object: MoneroPublicKey object
        """
        return self.m_pub_skey

    def PublicViewKey(self) -> MoneroPublicKey:
        """
        Return the public view key.

        Returns:
            MoneroPublicKey object: MoneroPublicKey object
        """
        return self.m_pub_vkey

    @lru_cache()
    def IntegratedAddress(self,
                          payment_id: bytes) -> str:
        """
        Return the integrated address with the specified payment ID.

        Args:
            payment_id (bytes): Payment ID

        Returns:
            str: Integrated address string
        """
        return XmrIntegratedAddrEncoder.EncodeKey(self.m_pub_skey.KeyObject(),
                                                  pub_vkey=self.m_pub_vkey.KeyObject(),
                                                  net_ver=self.m_coin_conf.IntegratedAddrNetVersion(),
                                                  payment_id=payment_id)

    @lru_cache()
    def PrimaryAddress(self) -> str:
        """
        Return the primary address.

        Returns:
            str: Primary address string
        """
        return self.m_subaddr.ComputeAndEncodeKeys(0,
                                                   0,
                                                   self.m_coin_conf.AddrNetVersion())

    @lru_cache()
    def Subaddress(self,
                   minor_idx: int,
                   major_idx: int = 0) -> str:
        """
        Return the specified subaddress.

        Args:
            minor_idx (int)          : Minor index (i.e. subaddress index)
            major_idx (int, optional): Major index (i.e. account index, default: 0)

        Returns:
            str: Subaddress string

        Raises:
            ValueError: If one of the indexes is not valid
        """
        if minor_idx == 0 and major_idx == 0:
            return self.PrimaryAddress()

        return self.m_subaddr.ComputeAndEncodeKeys(minor_idx,
                                                   major_idx,
                                                   self.m_coin_conf.SubaddrNetVersion())

    @staticmethod
    def __ViewFromSpendKey(priv_skey: MoneroPrivateKey) -> MoneroPrivateKey:
        """
        Get the private view key from the private spend key.

        Args:
            priv_skey (MoneroPrivateKey object): Private spend key

        Returns:
            MoneroPrivateKey object: Private view key
        """
        priv_vkey_bytes = Ed25519Utils.ScalarReduce(
            Kekkak256.QuickDigest(priv_skey.Raw().ToBytes())
        )
        return MoneroPrivateKey.FromBytes(priv_vkey_bytes)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for derivation scheme based on ed25519-blake2b curve as defined by BIP32 SLIP-0010."""

# Imports
from bip_utils.bip.bip32.slip10.bip32_slip10_ed25519 import Bip32Slip10Ed25519
from bip_utils.ecc import EllipticCurveTypes


class Bip32Slip10Ed25519Blake2b(Bip32Slip10Ed25519):
    """
    BIP32 SLIP-0010 ed25519-blake2b class.
    It allows master keys generation and keys derivation using ed25519-blake2b curve.
    """

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Return the elliptic curve type.

        Returns:
            EllipticCurveTypes: Curve type
        """
        return EllipticCurveTypes.ED25519_BLAKE2B


# Deprecated: only for compatibility
Bip32Ed25519Blake2bSlip = Bip32Slip10Ed25519Blake2b



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for Monero address encoding/decoding."""

# Imports
from typing import Any, Optional, Union

from bip_utils.addr.addr_dec_utils import AddrDecUtils
from bip_utils.addr.addr_key_validator import AddrKeyValidator
from bip_utils.addr.iaddr_decoder import IAddrDecoder
from bip_utils.addr.iaddr_encoder import IAddrEncoder
from bip_utils.base58 import Base58XmrDecoder, Base58XmrEncoder
from bip_utils.ecc import Ed25519MoneroPublicKey, IPublicKey
from bip_utils.utils.crypto import Kekkak256
from bip_utils.utils.misc import BytesUtils


class XmrAddrConst:
    """Class container for Monero address constants."""

    # Checksum length in bytes
    CHECKSUM_BYTE_LEN: int = 4
    # Payment ID length in bytes
    PAYMENT_ID_BYTE_LEN: int = 8


class _XmrAddrUtils:
    """Class container for Monero address utility functions."""

    @staticmethod
    def ComputeChecksum(payload_bytes: bytes) -> bytes:
        """
        Compute checksum in EOS format.

        Args:
            payload_bytes (bytes): Payload bytes

        Returns:
            bytes: Computed checksum
        """
        return Kekkak256.QuickDigest(payload_bytes)[:XmrAddrConst.CHECKSUM_BYTE_LEN]

    @staticmethod
    def DecodeAddr(addr: str,
                   net_ver_bytes: bytes,
                   payment_id_bytes: Optional[bytes] = None) -> bytes:
        """
        Decode a Monero address to bytes.

        Args:
            addr (str)                       : Address string
            net_ver_bytes (bytes)            : Net version
           payment_id_bytes (bytes, optional): Payment ID (only for integrated addresses)

        Returns:
            bytes: Public spend (first) and view (second) keys joined together

        Raises:
            ValueError: If the address encoding is not valid
        """

        # Decode from base58 XMR
        addr_dec_bytes = Base58XmrDecoder.Decode(addr)
        # Validate, remove prefix and split
        payload_bytes, checksum_bytes = AddrDecUtils.SplitPartsByChecksum(addr_dec_bytes,
                                                                          XmrAddrConst.CHECKSUM_BYTE_LEN)
        # Validate checksum
        AddrDecUtils.ValidateChecksum(payload_bytes, checksum_bytes, _XmrAddrUtils.ComputeChecksum)
        # Validate and remove prefix
        payload_bytes = AddrDecUtils.ValidateAndRemovePrefix(payload_bytes, net_ver_bytes)

        try:
            # Validate length without payment ID
            AddrDecUtils.ValidateLength(payload_bytes,
                                        Ed25519MoneroPublicKey.CompressedLength() * 2)
        except ValueError as ex:
            # Validate length with payment ID
            AddrDecUtils.ValidateLength(
                payload_bytes,
                (Ed25519MoneroPublicKey.CompressedLength() * 2) + XmrAddrConst.PAYMENT_ID_BYTE_LEN
            )
            # Check payment ID
            if payment_id_bytes is None or len(payment_id_bytes) != XmrAddrConst.PAYMENT_ID_BYTE_LEN:
                raise ValueError("Invalid payment ID") from ex

            payment_id_got_bytes = payload_bytes[-XmrAddrConst.PAYMENT_ID_BYTE_LEN:]
            if payment_id_bytes != payment_id_got_bytes:
                raise ValueError(f"Invalid payment ID (expected {BytesUtils.ToHexString(payment_id_bytes)}, "
                                 f"got {BytesUtils.ToHexString(payment_id_got_bytes)})") from ex

        # Validate public spend key
        pub_spend_key_bytes = payload_bytes[:Ed25519MoneroPublicKey.CompressedLength()]
        AddrDecUtils.ValidatePubKey(pub_spend_key_bytes, Ed25519MoneroPublicKey)
        # Validate public view key
        pub_view_key_bytes = payload_bytes[Ed25519MoneroPublicKey.CompressedLength():
                                           Ed25519MoneroPublicKey.CompressedLength() * 2]
        AddrDecUtils.ValidatePubKey(pub_view_key_bytes, Ed25519MoneroPublicKey)

        return pub_spend_key_bytes + pub_view_key_bytes

    @staticmethod
    def EncodeKey(pub_skey: Union[bytes, IPublicKey],
                  pub_vkey: Union[bytes, IPublicKey],
                  net_ver_bytes: bytes,
                  payment_id_bytes: Optional[bytes] = None) -> str:
        """
        Encode a public key to Monero address.

        Args:
            pub_skey (bytes or IPublicKey)    : Public spend key bytes or object
            pub_vkey (bytes or IPublicKey)    : Public view key bytes or object
            net_ver_bytes (bytes)             : Net version
            payment_id_bytes (bytes, optional): Payment ID (only for integrated addresses)

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519-monero
        """
        if payment_id_bytes is not None and len(payment_id_bytes) != XmrAddrConst.PAYMENT_ID_BYTE_LEN:
            raise ValueError("Invalid payment ID length")

        payment_id_bytes = b"" if payment_id_bytes is None else payment_id_bytes
        pub_spend_key_obj = AddrKeyValidator.ValidateAndGetEd25519MoneroKey(pub_skey)
        pub_view_key_obj = AddrKeyValidator.ValidateAndGetEd25519MoneroKey(pub_vkey)

        payload_bytes = (net_ver_bytes
                         + pub_spend_key_obj.RawCompressed().ToBytes()
                         + pub_view_key_obj.RawCompressed().ToBytes()
                         + payment_id_bytes)

        return Base58XmrEncoder.Encode(payload_bytes + _XmrAddrUtils.ComputeChecksum(payload_bytes))


class XmrAddrDecoder(IAddrDecoder):
    """
    Monero address decoder class.
    It allows the Monero address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Monero address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            net_ver (bytes): Expected net version

        Returns:
            bytes: Public spend (first) and view (second) keys joined together

        Raises:
            ValueError: If the address encoding is not valid
        """
        net_ver = kwargs["net_ver"]
        return _XmrAddrUtils.DecodeAddr(addr, net_ver)


class XmrAddrEncoder(IAddrEncoder):
    """
    Monero address encoder class.
    It allows the Monero address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Monero format.

        Args:
            pub_key (bytes or IPublicKey): Public spend key bytes or object

        Other Parameters:
            pub_vkey (bytes or IPublicKey): Public view key bytes or object
            net_ver (bytes)               : Net version
            payment_id (bytes, optional)  : Payment ID (only for integrated addresses)

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519-monero
        """
        pub_vkey = kwargs["pub_vkey"]
        net_ver = kwargs["net_ver"]

        return _XmrAddrUtils.EncodeKey(pub_key, pub_vkey, net_ver)


class XmrIntegratedAddrDecoder(IAddrDecoder):
    """
    Monero integrated address decoder class.
    It allows the Monero integrated address decoding.
    """

    @staticmethod
    def DecodeAddr(addr: str,
                   **kwargs: Any) -> bytes:
        """
        Decode a Monero address to bytes.

        Args:
            addr (str): Address string

        Other Parameters:
            net_ver (bytes)   : Expected net version
            payment_id (bytes): Expected payment ID

        Returns:
            bytes: Public spend (first) and view (second) keys joined together

        Raises:
            ValueError: If the address encoding is not valid
        """
        net_ver = kwargs["net_ver"]
        payment_id = kwargs["payment_id"]
        return _XmrAddrUtils.DecodeAddr(addr, net_ver, payment_id)


class XmrIntegratedAddrEncoder(IAddrEncoder):
    """
    Monero integrated address encoder class.
    It allows the Monero integrated address encoding.
    """

    @staticmethod
    def EncodeKey(pub_key: Union[bytes, IPublicKey],
                  **kwargs: Any) -> str:
        """
        Encode a public key to Monero integrated address.

        Args:
            pub_key (bytes or IPublicKey): Public spend key bytes or object

        Other Parameters:
            pub_vkey (bytes or IPublicKey): Public view key bytes or object
            net_ver (bytes)               : Net version
            payment_id (bytes)            : Payment ID

        Returns:
            str: Address string

        Raises:
            ValueError: If the public key is not valid
            TypeError: If the public key is not ed25519-monero
        """
        pub_vkey = kwargs["pub_vkey"]
        net_ver = kwargs["net_ver"]
        payment_id = kwargs["payment_id"]

        return _XmrAddrUtils.EncodeKey(pub_key, pub_vkey, net_ver, payment_id)


# Deprecated: only for compatibility, Encoder classes shall be used instead
XmrAddr = XmrAddrEncoder
XmrIntegratedAddr = XmrIntegratedAddrEncoder



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for Cardano Byron legacy master key generation.

References:
    https://input-output-hk.github.io/cardano-wallet/concepts/master-key-generation
    https://cips.cardano.org/cips/cip3/byron.md
"""

# Imports
from typing import Tuple

import cbor2

from bip_utils.bip.bip32 import IBip32MstKeyGenerator
from bip_utils.utils.crypto import HmacSha512, Sha512
from bip_utils.utils.misc import BitUtils


class CardanoByronLegacyMstKeyGeneratorConst:
    """Class container for Cardano Byron legacy BIP32 constants."""

    # HMAC message format
    HMAC_MESSAGE_FORMAT: bytes = b"Root Seed Chain %d"
    # Length in bytes for seed
    SEED_BYTE_LEN: int = 32


class CardanoByronLegacyMstKeyGenerator(IBip32MstKeyGenerator):
    """
    Cardano Byron legacy master key generator class.
    It allows master keys generation in according to Cardano Byron (legacy, used by old versions of Daedalus).
    """

    @classmethod
    def GenerateFromSeed(cls,
                         seed_bytes: bytes) -> Tuple[bytes, bytes]:
        """
        Generate a master key from the specified seed.

        Args:
            seed_bytes (bytes): Seed bytes

        Returns:
            tuple[bytes, bytes]: Private key bytes (index 0) and chain code bytes (index 1)

        Raises:
            Bip32KeyError: If the seed is not suitable for master key generation
            ValueError: If seed length is not valid
        """
        if len(seed_bytes) != CardanoByronLegacyMstKeyGeneratorConst.SEED_BYTE_LEN:
            raise ValueError(f"Invalid seed length ({len(seed_bytes)})")
        return cls.__HashRepeatedly(cbor2.dumps(seed_bytes), 1)

    @classmethod
    def __HashRepeatedly(cls,
                         data_bytes: bytes,
                         itr_num: int) -> Tuple[bytes, bytes]:
        """
        Continue to hash the data bytes until the third-highest bit of the last byte is not zero.

        Args:
            data_bytes (bytes): Data bytes
            itr_num (int)     : Iteration number

        Returns:
            tuple[bytes, bytes]: Key bytes (index 0) and chain code bytes (index 1)
        """
        il_bytes, ir_bytes = HmacSha512.QuickDigestHalves(
            data_bytes,
            CardanoByronLegacyMstKeyGeneratorConst.HMAC_MESSAGE_FORMAT % itr_num
        )
        key_bytes = cls.__TweakMasterKeyBits(Sha512.QuickDigest(il_bytes))
        if BitUtils.AreBitsSet(key_bytes[31], 0x20):
            return cls.__HashRepeatedly(data_bytes, itr_num + 1)
        return key_bytes, ir_bytes

    @staticmethod
    def __TweakMasterKeyBits(key_bytes: bytes) -> bytes:
        """
        Tweak master key bits.

        Args:
            key_bytes (bytes): Key bytes

        Returns:
            bytes: Tweaked key bytes
        """
        key_bytes = bytearray(key_bytes)
        # Clear the lowest 3 bits of the first byte of kL
        key_bytes[0] = BitUtils.ResetBits(key_bytes[0], 0x07)
        # Clear the highest bit of the last byte of kL
        key_bytes[31] = BitUtils.ResetBits(key_bytes[31], 0x80)
        # Set the second-highest bit of the last byte of kL
        key_bytes[31] = BitUtils.SetBits(key_bytes[31], 0x40)

        return bytes(key_bytes)
